# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import re
from pathlib import Path

import pytest

from fedrq._archive import RPMArchive, RPMArchiveError


def test_archive_extract_specfile(tmp_path: Path, data_path: Path):
    repo1 = data_path / "repos/repo1/"
    srpm = repo1 / "repo/SRPMS/specs/packageb-1-1.fc36.src.rpm"
    with RPMArchive(srpm) as archive:
        archive.extract_specfile(tmp_path)

    expected_spec = repo1 / "specs/packageb.spec"
    gotten = tmp_path / "packageb.spec"
    assert expected_spec.read_bytes() == gotten.read_bytes()


def test_archive_extract_specfile_error(
    tmp_path: Path, data_path: Path, target_cpu: str
):
    repo1 = data_path / "repos/repo1/"
    invalid = repo1 / f"repo/RPMS/specs/packageb-1-1.fc36.{target_cpu}.rpm"
    assert invalid.is_file()
    dest = tmp_path / "abc"
    dest.mkdir()
    with RPMArchive(invalid) as archive:  # noqa SIM177
        with pytest.raises(
            RPMArchiveError, match=re.escape(f"{archive} is not a source rpm")
        ):
            archive.extract_specfile(dest)
    assert next(dest.iterdir(), None) is None



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
download commands
"""

from __future__ import annotations

import argparse
import logging
import shutil
import sys
from collections.abc import Callable, Iterator
from contextlib import suppress
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import requests
import rpm

from fedrq._archive import RPMArchive
from fedrq._utils import exhaust_it
from fedrq.backends.base import PackageCompat
from fedrq.cli.base import Command

_LOG = logging.getLogger(__name__)


def callback_ind(start: int, end: int) -> Callable[[Path], None]:
    def inner(destpath: Path) -> None:
        print(f"Downloading {destpath.name} ({start}/{end})")

    return inner


def _get_req_kwargs(package: PackageCompat) -> dict[str, Any]:  # pragma: no cover
    req_kwargs: dict[str, Any] = {}
    repo_obj = package.repo
    # libdnf5 has a .get_config() attribute
    with suppress(AttributeError):
        repo_obj = repo_obj.get_config()
    if repo_obj.sslverify is False:
        repo_obj["verify"] = False
    elif repo_obj.sslcacert:
        req_kwargs["verify"] = repo_obj.sslcacert
    if repo_obj.sslclientcert and repo_obj.sslclientkey:
        req_kwargs["cert"] = (repo_obj.sslclientcert, repo_obj.sslclientkey)
    return req_kwargs


def download(
    package: PackageCompat,
    destdir: Path,
    callback: Callable[[Path], None] | None = None,
) -> Path:
    def _remote_dl(url: str, src: Path, dest: Path) -> None:  # noqa: ARG001
        req_kwargs = _get_req_kwargs(package)
        req = requests.get(url, allow_redirects=True, **req_kwargs)
        req.raise_for_status()
        dest.write_bytes(req.content)

    def _local_cp(url: str, src: Path, dest: Path) -> None:  # noqa: ARG001
        shutil.copy2(src, dest)

    url = package.remote_location()
    if not url:  # pragma: no cover
        raise ValueError(f"Could not determine url for {package}!")

    mode: Callable[[str, Path, Path], None]
    src: Path
    dest: Path

    if url.startswith("file://"):
        src = Path(url[7:])
        mode = _local_cp
    else:
        src = Path(url)
        mode = _remote_dl

    dest = destdir / src.name
    if callback:
        callback(dest)
    mode(url, src, dest)
    return dest


class DownloadCommand(Command):
    """
    Download an (S)RPM from the repos.
    No gpg checking is preformed.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        parser = super().make_parser(
            parser_func,
            add_help=add_help,
            help=(cls.__doc__ or "").strip().replace("\n", " "),
            parents=[
                cls.parent_parser(),
                cls.arch_parser(),
                cls.resolve_parser(),
                cls.assume_parser(),
            ],
            **kwargs,
        )
        parser.add_argument("-o", "--destdir", type=Path, default=Path())
        return parser

    def _select_packages(self) -> None:
        self.query = self.rq.resolve_pkg_specs(
            self.args.names, self.args.resolve_packages, self.args.latest
        )
        self.rq.arch_filterm(self.query, self.args.arch)

    def _prompt(self) -> bool:
        if not self.query:
            sys.exit("No packages found!")
        size = len(self.query)
        inflected = "package"
        if size > 1:
            inflected += "s"
        print(f"\nDownload {size} {inflected}")

        if self.args.assumeyes:
            return True
        if self.args.dry_run:
            return False

        inp = input("(Y/n) ")
        if inp.lower() in ("", "y"):
            return True
        print("Exiting...", file=sys.stderr)
        return False

    def _downloadit(self, destdir: Path) -> Iterator[tuple[PackageCompat, Path]]:
        size = len(self.query)
        for ind, package in enumerate(self.query):
            try:
                dest = download(package, destdir, callback_ind(ind + 1, size))
                yield package, dest
            except Exception as exc:
                _LOG.debug("Handled error:", exc_info=exc)
                sys.exit(f"Failed to download {package}: {exc!r}")

    def run(self) -> None:
        self._select_packages()
        for line in self.format():
            print(line)
        if not self._prompt():
            return
        exhaust_it(self._downloadit(self.args.destdir))


class DownloadSpecCommand(DownloadCommand):
    """
    Download an SRPM and extract its specfile.
    No gpg checking is preformed.
    """

    @classmethod
    def make_parser(
        cls,
        parser_func: Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        parser = super(DownloadCommand, cls).make_parser(
            parser_func,
            help=(cls.__doc__ or "").strip().replace("\n", " "),
            add_help=add_help,
            parents=[
                cls.parent_parser(),
                cls.assume_parser(),
            ],
            **kwargs,
        )
        parser.add_argument("-o", "--destdir", type=Path, default=Path())
        parser.add_argument(
            "--gpgcheck",
            action=argparse.BooleanOptionalAction,
            default=False,
            help="Whether to check package gpg keys."
            " The keys must be part of the system RPM keyring."
            " Default: --no-gpgcheck",
        )
        parser.set_defaults(resolve_packages=False, arch="src")
        return parser

    def run(self) -> None:
        self._select_packages()
        for line in self.format():
            print(line)
        if not self._prompt():
            return

        ts = rpm.TransactionSet()
        if not self.args.gpgcheck:
            ts.setVSFlags(rpm.RPMVSF_MASK_NOSIGNATURES)
        with TemporaryDirectory() as _tmp:
            temp = Path(_tmp)
            for _, dest in self._downloadit(temp):
                with RPMArchive(dest, ts=ts) as archive:
                    archive.extract_specfile(self.args.destdir)
                dest.unlink()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
PRIVATE: Utilities for dealing with RPM archives
"""

from __future__ import annotations

from collections.abc import Iterator
from contextlib import closing, contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import rpm

if TYPE_CHECKING:
    from typing import TypeVar

    _ArchiveT = TypeVar("_ArchiveT", bound="RPMArchive")


class RPMArchiveError(Exception):
    """
    Error occured while handling an RPMArchive
    """


class RPMArchive:
    def __init__(self, path: Path, *, ts: rpm.TransactionSet | None = None) -> None:
        self.path = path
        self.ts: rpm.TransactionSet = ts or rpm.TransactionSet()
        self.fd: rpm.fd = rpm.fd(str(path), "r")
        self.hdr: rpm.hdr = self.ts.hdrFromFdno(self.fd)
        self._loc: int = self.fd.tell()

    def __repr__(self) -> str:  # pragma: no cover
        path, ts = self.path, self.ts
        typ = type(self).__name__
        return f"{typ}({path=}, {ts=})"

    @contextmanager
    def unpack_archive(self) -> Iterator[tuple[rpm.files, rpm.fd, rpm.archive]]:
        """
        Open the RPM's archive payload.

        Yields:
            A tuple of the `rpm.files` object, the payload's open `rpm.fd`, and
            the `rpm.archive` object
        """
        self.fd.seek(self._loc)
        files: rpm.files = rpm.files(self.hdr)
        payload: rpm.fd = rpm.fd(self.fd, "r", self.hdr["payloadcompressor"])
        archive: rpm.archive = files.archive(payload)
        with closing(payload), closing(archive):
            yield files, payload, archive

    def _extract_afile(
        self, archive: rpm.archive, file: rpm.file, destdir: Path
    ) -> Path:
        final = destdir / file.name
        with closing(rpm.fd(str(final), "w")) as fd:
            archive.readto(fd)
        return final

    def extract_specfile(self, destdir: Path) -> Path:
        """
        Extract the SRPM's specfile to `destdir`.

        Args:
            destdir:
                Directory in which to extract the specfile.
                The directory must already exist
        Raises:
            `RPMArchiveError` if `self` is not a *source* RPM

        """
        if not self.hdr.isSource():
            raise RPMArchiveError(f"{self} is not a source rpm")
        with self.unpack_archive() as (_, _, archive):
            for afile in archive:
                if afile.fflags & rpm.RPMFILE_SPECFILE:
                    return self._extract_afile(archive, afile, destdir)
        raise RPMArchiveError("No specfile found in the SRPM")  # pragma: no cover

    def __enter__(self: _ArchiveT) -> _ArchiveT:
        return self

    def __exit__(self, *_) -> None:
        self.fd.close()


__all__ = ("RPMArchive", "RPMArchiveError")



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from collections.abc import Callable, Collection, Sequence
from typing import Any

import pytest
from pytest_mock import MockerFixture

import fedrq.cli
from fedrq.cli.formatters import DefaultFormatters


def test_formatters_command(capsys: pytest.CaptureFixture) -> None:
    fedrq.cli.main(["formatters"])
    expected_out = """\
a
arch
attr:
buildtime
conflicts
debug_name
description
downloadsize
e
enhances
epoch
evr
files
from_repo
full_nevra
installsize
installtime
json:
license
line:
location
multiline:
na
na-requiresmatch-src:
na-requiresmatch:
name
narm:
narmsrc:
nev
nevr
nevra
nevrr
nv
obsoletes
packager
plain
plainwithrepo
provides
r
reason
recommends
release
remote_location
repoid
reponame
requires
requiresmatch-src:
requiresmatch:
rm:
rmsrc:
size
source
source+requiresmatch:
source+rm:
source_debug_name
source_name
sourcerpm
src
suggests
summary
supplements
url
v
vendor
version
"""
    out, err = capsys.readouterr()
    assert not err
    assert out == expected_out


@pytest.mark.parametrize(
    "args,formatters_kwargs,output_check",
    [
        pytest.param(
            [], {"attrs": True, "formatters": True, "special_formatters": True}, None
        ),
        pytest.param(
            ["--formatters"],
            {"attrs": False, "special_formatters": False, "formatters": True},
            None,
        ),
        pytest.param(
            ["--formatters", "--attrs"],
            {"attrs": True, "special_formatters": False, "formatters": True},
            None,
        ),
        pytest.param(
            ["--special-formatters"],
            {"attrs": False, "special_formatters": True, "formatters": False},
            lambda items: all(item[-1] == ":" for item in items),
        ),
    ],
)
def test_formatters_command_options(
    mocker: MockerFixture,
    args: Sequence[str],
    formatters_kwargs: dict[str, bool],
    output_check: Callable[[Collection[str]], Any] | None,
    capsys: pytest.CaptureFixture,
) -> None:
    mocked = mocker.spy(DefaultFormatters.__class__, "formatters_it")
    fedrq.cli.main(["formatters", *args])
    mocked.assert_called_once_with(DefaultFormatters, **formatters_kwargs)
    out, err = capsys.readouterr()
    assert not err
    if output_check:
        output_check(out.splitlines())



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import datetime
from collections.abc import Sequence

import pytest

import fedrq.cli
from fedrq.backends.base import ChangelogEntry, RepoqueryBase
from fedrq.cli.commands.changelogs import _positive_int


def test_get_changelogs(repo_test_rq: RepoqueryBase):
    query = repo_test_rq.query()
    for package in query:
        evr = package.evr.split(".", 1)[0]
        expected = ChangelogEntry(
            text="- Dummy changelog entry",
            author=f"Maxwell G <maxwell@gtmx.me> - {evr}",
            date=datetime.date(2023, 7, 16),
        )
        entries = list(repo_test_rq.backend.get_changelogs(package))
        assert entries == [expected]


@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["packagea"]),
        pytest.param(["--src", "packagea"]),
        pytest.param(["--notsrc", "packagea"]),
        pytest.param(["--entry-limit=1", "packagea"]),
    ],
)
def test_changelog_cmd(
    patch_config_dirs, args: Sequence[str], capsys: pytest.CaptureFixture
):
    fedrq.cli.main(["changelog", *args])
    expected = """\
* Sun Jul 16 2023 Maxwell G <maxwell@gtmx.me> - 1-1
- Dummy changelog entry

"""
    out, _ = capsys.readouterr()
    assert out == expected


@pytest.mark.parametrize(
    "args, expected_error",
    [
        pytest.param(["fjsfasdfa"], "No matches for package!"),
    ],
)
def test_changelog_cmd_error_cases(
    patch_config_dirs, args: Sequence[str], expected_error: str
):
    with pytest.raises(SystemExit, match=expected_error):
        fedrq.cli.main(["changelog", *args])


def test_changelog_cmd_too_many_names(patch_config_dirs, capsys: pytest.CaptureFixture):
    with pytest.raises(SystemExit, match="^1$"):
        fedrq.cli.main(["changelog", "packagea", "packageb"])
    assert capsys.readouterr()[1] == "ERROR: More than one package name was passed!\n"


def test_changelog_positive_int_error():
    with pytest.raises(TypeError, match="--entry-limit must be positive!"):
        _positive_int(str(-5))



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections import abc as cabc

from fedrq._utils import filter_latest
from fedrq.cli import Command


class Pkgs(Command):
    """
    Find the packages that match NAMES.
    NAMES can be package package name globs or NEVRs.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        parser = super().make_parser(
            parser_func,
            parents=[cls.parent_parser(), cls.arch_parser(), cls.resolve_parser()],
            help="Find the packages that match a list of package specs",
            add_help=add_help,
            **kwargs,
        )

        return parser

    def run(self) -> None:
        self.query = self.rq.query(empty=True)

        resolved_packages = self.rq.resolve_pkg_specs(
            self.args.names, self.args.resolve_packages
        )
        self._logq(resolved_packages, "resolved_packages")
        self.query = self.query.union(resolved_packages)

        glob_packages = self.rq.query(name__glob=self.args.names)
        self._logq(glob_packages, "glob_packages")
        self.query = self.query.union(glob_packages)

        self.query = self.rq.arch_filter(self.query, self.args.arch)
        filter_latest(self.query, self.args.latest)
        self._logq(self.query, "self.query")

        for p in self.format():
            print(p)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This module contains a fedrq backend (i.e. an implementation of the
[`fedrq.backends.base.BackendMod`][fedrq.backends.base.BackendMod] interface)
that uses the dnf Python bindings.
"""

from __future__ import annotations

import logging
import sys
import typing as t
from collections.abc import Collection, Iterator
from enum import Enum
from functools import cache

from fedrq._utils import filter_latest
from fedrq.backends import MissingBackendError
from fedrq.backends.base import (
    BackendMod,
    BaseMakerBase,
    ChangelogEntry,
    PackageCompat,
    PackageQueryCompat,
    RepoqueryBase,
)
from fedrq.backends.dnf import BACKEND

try:
    import dnf
    import hawkey
except ImportError:
    raise MissingBackendError from None

if t.TYPE_CHECKING:
    from _typeshed import StrPath

LOG = logging.getLogger(__name__)

PackageCompat.register(dnf.package.Package)
Package: type[PackageCompat] = dnf.package.Package
PackageQueryCompat.register(dnf.query.Query)
PackageQuery: type[PackageQueryCompat] = dnf.query.Query
RepoError = dnf.exceptions.RepoError


class BaseMaker(BaseMakerBase):
    """
    Create a Base object and load repos
    """

    base: dnf.Base

    def __init__(self, base: dnf.Base | None = None) -> None:
        """
        Initialize and configure the base object.
        """
        self.base: dnf.Base = base or dnf.Base()

    @property
    def conf(self) -> dnf.conf.MainConf:
        return self.base.conf

    def set(self, key: str, value: t.Any) -> None:
        setattr(self.conf, key, value)

    def set_var(self, key: str, value: t.Any) -> None:
        if key not in self.base.conf.substitutions:
            raise KeyError(f"{key} is not a valid substitution")
        self.set(key, value)

    def load_filelists(self, enable: bool = True) -> None:
        # Old versions of dnf always load filelists
        try:
            types: list[str] = self.conf.optional_metadata_types
        except AttributeError:
            return
        if enable:
            types.append("filelists")
            return
        while "filelists" in types:
            types.remove("filelists")

    def load_changelogs(self, enable: bool = True) -> None:
        for repo in self.base.repos.iter_enabled():
            repo.load_metadata_other = enable

    def fill_sack(
        self,
        *,
        from_cache: bool = False,
        load_system_repo: bool = False,
    ) -> dnf.Base:
        """
        Fill the sack and returns the dnf.Base object.
        The repository configuration shouldn't be manipulated after this.

        Note that the `_cachedir` arg is private and subject to removal.
        """
        if from_cache:
            self.base.fill_sack_from_repos_in_cache(load_system_repo=load_system_repo)
        else:
            self.base.fill_sack(load_system_repo=load_system_repo)
        return self.base

    def read_system_repos(self, disable: bool = True) -> None:
        """
        Load system repositories into the base object.
        By default, they are all disabled even if 'enabled=1' is in the
        repository configuration.
        """
        self.base.read_all_repos()
        if not disable:
            return None
        for repo in self.base.repos.iter_enabled():
            repo.disable()

    def enable_repos(self, repos: Collection[str]) -> None:
        """
        Enable a list of repositories by their repoid.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        for repo in repos:
            self.enable_repo(repo)

    def enable_repo(self, repo: str) -> None:
        """
        Enable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        if repo_obj := self.base.repos.get_matching(repo):
            repo_obj.enable()
        else:
            raise ValueError(f"{repo} repo definition was not found.")

    def disable_repo(self, repo: str, ignore_missing: bool = True) -> None:
        """
        Disable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration
        when ignore_missing is False.
        """
        if repo_obj := self.base.repos.get_matching(repo):
            repo_obj.disable()
        elif not ignore_missing:
            raise ValueError(f"{repo} repo definition was not found.")

    def read_repofile(self, file: StrPath) -> None:
        rr = dnf.conf.read.RepoReader(self.base.conf, None)
        for repo in rr._get_repos(str(file)):
            LOG.debug("Adding %s from %s", repo.id, file)
            self.base.repos.add(repo)

    # This is private for now
    def _read_repofile_new(self, file: StrPath, ensure_enabled: bool = False) -> None:
        """
        Load repositories from a repo file if they're not already in the
        configuration.
        """
        rr = dnf.conf.read.RepoReader(self.base.conf, None)
        for repo in rr._get_repos(str(file)):
            if repo.id in self.base.repos:
                LOG.debug("Not adding %s. It's already in the config.", repo.id)
            else:
                LOG.debug("Adding %s from %s", repo.id, file)
                self.base.repos.add(repo)
            if ensure_enabled:
                LOG.debug("Ensuring that %s is enabled.", repo.id)
                self.base.repos[repo.id].enable()

    def create_repo(self, repoid: str, **kwargs: t.Any) -> None:
        """
        Add a Repo object to the repo sack and configure it.

        Args:
            kwargs:
                key-values options that should be set on the Repo object values
                (like `$basearch`) will be substituted automatically.
        """
        self.base.repos.add_new_repo(repoid, self.conf, **kwargs)

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]

    def repolist(self, enabled: bool | None = None) -> list[str]:
        if enabled is None:
            return list(self.base.repos)
        return [r.id for r in self.base.repos.values() if r.enabled is bool(enabled)]

    def enable_source_repos(self) -> None:
        self.base.repos.enable_source_repos()


class NEVRAForms(int, Enum):
    NEVRA = hawkey.FORM_NEVRA
    NEVR = hawkey.FORM_NEVR
    NEV = hawkey.FORM_NEV
    NA = hawkey.FORM_NA
    NAME = hawkey.FORM_NAME


# Use PackageCompat and PackageQueryCompat as the TypeVar, as the the native dnf objects
# don't provide typing.
class Repoquery(RepoqueryBase[PackageQueryCompat[PackageCompat]]):
    def __init__(self, base: dnf.Base) -> None:
        self.base: dnf.Base = base

    @property
    def base_arches(self) -> set[str]:
        return {self.base.conf.arch, self.base.conf.basearch}

    def _query(self) -> dnf.query.Query:
        return self.base.sack.query()

    def resolve_pkg_specs(
        self,
        specs: Collection[str],
        resolve: bool = False,
        latest: int | None = None,
        with_src: bool = True,
        *,
        with_filenames: bool | None = None,
        with_provides: bool | None = None,
        resolve_provides: bool | None = None,
        nevra_forms: list[NEVRAForms] | None = None,
    ) -> dnf.query.Query:
        opts = self._get_resolve_options(
            resolve, with_filenames, with_provides, resolve_provides
        )
        resolve_provides = opts.pop("resolve_provides")
        opts["with_src"] = with_src
        if nevra_forms:
            opts["forms"] = nevra_forms

        query = self.query(empty=True)
        for p in specs:
            subject = dnf.subject.Subject(p).get_best_query(self.base.sack, **opts)
            query = query.union(subject)
        if resolve_provides:
            query = query.union(self.query(provides=specs))
        filter_latest(query, latest)
        return query

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]


@cache
def get_releasever():
    """
    Return the system releasever
    """
    return dnf.rpm.detect_releasever("/")


def get_changelogs(package: t.Any) -> Iterator[ChangelogEntry]:
    for entry in package.changelogs:
        yield ChangelogEntry(
            text=entry["text"], author=entry["author"], date=entry["timestamp"]
        )


__all__ = (
    "BACKEND",
    "BaseMaker",
    "Package",
    "NEVRAForms",
    "PackageQuery",
    "RepoError",
    "Repoquery",
    "get_releasever",
    "get_changelogs",
    #
    "dnf",
    "hawkey",
)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later


from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Any

import pytest

from fedrq import config as rqconfig
from fedrq.backends.base import (
    BackendMod,
    PackageCompat,
    PackageQueryCompat,
    RepoqueryBase,
)


def test_make_base_rawhide_repos() -> None:
    config = rqconfig.get_config()
    rawhide = config.get_release("rawhide")
    bm = config.backend_mod.BaseMaker()
    base = rawhide.make_base(config, fill_sack=False, base_maker=bm)  # noqa: F841
    repos = bm.repolist(True)
    assert len(repos) == len(rawhide.repog.repos)
    assert set(repos) == set(rawhide.repog.repos)


def test_package_protocol(repo_test_rq: RepoqueryBase):
    package = repo_test_rq.get_package("packagea", arch="noarch")
    assert isinstance(package, PackageCompat)


def test_query_protocol(repo_test_rq: RepoqueryBase):
    query = repo_test_rq.query()
    assert isinstance(query, PackageQueryCompat)


def test_baseurl_repog(repo_test_rq: RepoqueryBase, data_path: Path):
    for i in ("", "file://"):
        second_rq = rqconfig.get_config().get_rq(
            "rawhide", f"@baseurl:{i}{data_path/ 'repos' / 'repo1' / 'repo'}"
        )
        assert sorted(map(str, repo_test_rq.query())) == sorted(
            map(str, second_rq.query())
        )


@pytest.mark.parametrize("explicit_forms", [pytest.param(True), pytest.param(False)])
def test_resolve_pkg_specs_forms(
    explicit_forms: bool, repo_test_rq: RepoqueryBase, default_backend: BackendMod
) -> None:
    packages = [repo_test_rq.get_package("packagea", arch="noarch")]
    assert len(packages) == 1
    packages2 = packages + [repo_test_rq.get_package("packagea", arch="src")]

    func = repo_test_rq.resolve_pkg_specs
    forms = default_backend.NEVRAForms

    # nevra_forms excludes "packagea.noarch" (0)
    nevra_forms = [forms.NAME]
    assert not list(func(["packagea.noarch"], nevra_forms=nevra_forms))

    # forms.NAME (2)
    nevra_forms = [forms.NAME] if explicit_forms else None
    assert list(func(["packagea"], nevra_forms=nevra_forms)) == packages2

    # forms.NEV (2)
    nevra_forms = [forms.NEV] if explicit_forms else None
    assert list(func(["packagea-0:1"], nevra_forms=nevra_forms)) == packages2

    # forms.NA (1)
    nevra_forms = [forms.NAME, forms.NA]
    assert list(func(["packagea.noarch"], nevra_forms=nevra_forms)) == packages

    # forms.NEVRA (1)
    nevra_forms = [forms.NEVRA]
    assert list(func(["packagea-1-1.fc36.noarch"], nevra_forms=nevra_forms)) == packages


@pytest.mark.parametrize(
    "specs,kwargs,count",
    [
        pytest.param(["/usr/share/packagea"], {"resolve": True}, 1, id="resolve"),
        pytest.param(
            ["/usr/share/packagea"], {"resolve": False}, 0, id="resolve-false"
        ),
        pytest.param(
            ["/usr/share/packagea"],
            {"resolve": True, "with_filenames": False},
            0,
            id="resolve-without-filenames",
        ),
        pytest.param(
            ["/usr/share/packagea", "package(a)"],
            {"resolve": True, "with_filenames": False},
            1,
            id="resolve-without-filenames-provides",
        ),
        pytest.param(
            ["/usr/share/packagea"],
            {"resolve": True, "with_filenames": True},
            1,
            id="resolve-with-explict-filenames",
        ),
        pytest.param(
            ["/usr/share/packagea"],
            {"with_filenames": True},
            1,
            id="with-filenames",
        ),
    ],
)
def test_resolve_pkg_specs_resolve(
    repo_test_rq: RepoqueryBase,
    specs: Sequence[str],
    kwargs: dict[str, Any],
    count: int,
) -> bool:
    results = repo_test_rq.resolve_pkg_specs(specs, **kwargs)
    assert len(results) == count



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
changelog commands
"""

from __future__ import annotations

import argparse
import sys

from fedrq.backends.base import PackageCompat, PackageQueryCompat
from fedrq.cli.base import Command, v_add_errors


def _get_single_package(query: PackageQueryCompat) -> PackageCompat:
    if not query:
        sys.exit("No matches for package!")
    arches = {package.arch for package in query}
    if "src" in arches:
        for package in query:
            if package.arch == "src":
                return package
    if len(arches) != 1:
        sys.exit("Found duplicate packages:", *query)
    return next(iter(query))


def _positive_int(number: str) -> int:
    i = int(number)
    if i < 1:
        raise TypeError("--entry-limit must be positive!")
    return i


class ChangelogCommand(Command):
    """
    Retrieve a single package's changelog
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.config.load_other_metadata = True
        self._v_names()
        self.v_default()

    @v_add_errors
    def _v_names(self) -> str | None:
        if len(self.args.names) > 1:
            return "More than one package name was passed!"
        return None

    @classmethod
    def make_parser(
        cls, parser_func=argparse.ArgumentParser, *, add_help: bool = False, **kwargs
    ) -> argparse.ArgumentParser:
        parser = super().make_parser(
            parser_func,
            add_help=add_help,
            parents=[
                cls.parent_parser(formatter=False, latest=False),
                cls.arch_parser(),
            ],
            **kwargs,
        )
        parser.add_argument("--entry-limit", type=_positive_int)
        return parser

    def run(self) -> None:
        query = self.rq.resolve_pkg_specs(self.args.names, latest=1)
        self.rq.arch_filterm(query, self.args.arch)
        changelogs = sorted(
            self.backend.get_changelogs(_get_single_package(query)),
            key=lambda entry: entry.date,
            reverse=True,
        )
        if not changelogs:  # pragma: no cover
            sys.exit("(empty)")
        for i, entry in enumerate(changelogs):
            print(entry)
            print()
            if self.args.entry_limit is not None and i + 1 >= self.args.entry_limit:
                break



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from pytest import CaptureFixture, mark

import fedrq.cli


@mark.no_rpm_mock
def test_cli_make_cache(capsys: CaptureFixture):
    fedrq.cli.main(["make-cache", "-b", "f38"])
    out, err = capsys.readouterr()
    assert out == "Loaded 4 repos\n"
    assert not err



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Test dnf-specific backend code
"""

from __future__ import annotations

import pytest

from fedrq.backends.base import BackendMod


@pytest.fixture(autouse=True)
def skip_mod(default_backend: BackendMod):
    if default_backend.BACKEND != "dnf":
        pytest.skip("This test checks libdnf5 functionality")


def test_bm_set_var_invalid(default_backend: BackendMod):
    bm = default_backend.BaseMaker()
    with pytest.raises(KeyError, match="best is not a valid substitution"):
        bm.set_var("best", "")



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Commands for managing the repo cache
"""

from __future__ import annotations

import argparse
from collections.abc import Callable

from fedrq.cli.base import Command


class MakeCacheCommand(Command):
    """
    Load the repodata for the current branch/repo config
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        parser = super().make_parser(
            parser_func,
            add_help=add_help,
            parents=[
                cls.parent_parser(formatter=False, latest=False, name=False),
                cls.arch_parser(),
            ],
            **kwargs,
        )
        return parser

    def run(self) -> None:
        enabled_repos = self.backend.BaseMaker(self.rq.base).repolist(True)
        inflected = "repo" if len(enabled_repos) == 1 else "repos"
        print(f"Loaded {len(enabled_repos)} {inflected}")



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Generic repoquery backend tests
"""

from __future__ import annotations

from fedrq.backends.base import BackendMod

BACKEND_MEMBERS: set[str] = set(BackendMod.__annotations__)


def test_repoquery_interface():
    """
    Ensure the fedrq.repoquery module wrapper implements the full backend
    interface
    """
    import fedrq.repoquery

    assert set(dir(fedrq.repoquery)) & BACKEND_MEMBERS == BACKEND_MEMBERS


def test_backend_interface(default_backend):
    """
    Ensure the current backend implements the full backend interface
    """

    assert set(dir(default_backend)) & BACKEND_MEMBERS == BACKEND_MEMBERS



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from pathlib import Path

from pytest import CaptureFixture

import fedrq.cli


def test_download_spec(
    patch_config_dirs, tmp_path: Path, data_path: Path, capsys: CaptureFixture
):
    tmp_path /= "abc"
    tmp_path.mkdir()
    repo1 = data_path / "repos/repo1/"
    # srpm = repo1 / "repo/SRPMS/specs/packagea-1-1.fc36.src.rpm"
    expected_spec = repo1 / "specs/packagea.spec"

    fedrq.cli.main(["download-spec", "packagea", "-o", str(tmp_path), "-y"])
    gotten = tmp_path / "packagea.spec"
    assert list(tmp_path.iterdir()) == [gotten]
    assert expected_spec.read_bytes() == gotten.read_bytes()


def test_download(
    patch_config_dirs, tmp_path: Path, data_path: Path, capsys: CaptureFixture
):
    tmp_path /= "xyz"
    tmp_path.mkdir()
    repo = data_path / "repos/repo1/repo"
    expected = [
        repo / "SRPMS/specs/packagea-1-1.fc36.src.rpm",
        repo / "RPMS/specs/packagea-1-1.fc36.noarch.rpm",
    ]

    fedrq.cli.main(["download", "packagea", "-o", str(tmp_path), "-y"])
    gotten = tmp_path / "packagea.spec"
    assert len(list(tmp_path.iterdir())) == 2
    for expect in expected:
        gotten = tmp_path / expect.name
        assert expect.read_bytes() == gotten.read_bytes()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import pytest

import fedrq.config
from fedrq.backends.base import BackendMod, RepoqueryBase


def test_load_changelogs(default_backend: BackendMod):
    bm = default_backend.BaseMaker()
    config = fedrq.config.get_config()
    release = config.get_release("f37")
    bm.load_release_repos(release)
    expected_repos = [
        "fedora",
        "fedora-source",
        "updates",
        "updates-source",
    ]
    if default_backend.BACKEND == "dnf":
        bm.load_changelogs()
        enabled = list(bm.base.repos.iter_enabled())
        assert sorted(repo.id for repo in enabled) == expected_repos
        for key in expected_repos:
            assert bm.base.repos[key].load_metadata_other is True
        bm.load_changelogs(False)
        for key in expected_repos:
            assert bm.base.repos[key].load_metadata_other is False
    else:
        old = sorted(bm.base.get_config().optional_metadata_types)
        assert "other" not in old
        bm.load_changelogs(False)
        assert sorted(bm.base.get_config().optional_metadata_types) == old
        bm.load_changelogs()
        expected = sorted((*old, "other"))
        assert expected == sorted(bm.base.get_config().optional_metadata_types)
        bm.load_changelogs(False)
        assert sorted(bm.base.get_config().optional_metadata_types) == old


def test_bm_enable_disable(
    repo_test_rq: RepoqueryBase, default_backend: BackendMod
) -> None:
    bm = default_backend.BaseMaker(repo_test_rq.base)
    assert bm.repolist(True) == ["testrepo1"]
    bm.disable_repo("testrepo1")
    assert not bm.repolist(True)
    bm.enable_repos(["testrepo1"])

    with pytest.raises(
        ValueError, match="does-not-exist repo definition was not found."
    ):
        bm.disable_repo("does-not-exist", ignore_missing=False)


def test_bm_backend_property(default_backend: BackendMod) -> None:
    assert default_backend.BaseMaker().backend == default_backend



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from collections.abc import Iterator
from typing import Any

from fedrq._utils import exhaust_it


def fake_iter(lst: list[Any]) -> Iterator[None]:
    while lst:
        lst.pop()
        yield


def test_util_exhaust_it():
    lst = [*range(10)]
    exhaust_it(fake_iter(lst))
    assert not lst



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Generic backends tests
"""

from __future__ import annotations

import pytest

from fedrq.backends import get_default_backend
from fedrq.backends.base import BackendMod


def test_backends_no_allow_multiple_backends_per_process(
    default_backend: BackendMod,
) -> None:
    other_backends = {"dnf": "libdnf5", "libdnf5": "dnf"}
    other = other_backends[default_backend.BACKEND]
    with pytest.warns(
        UserWarning,
        match=f"Falling back to {default_backend.BACKEND}. {other} cannot be used.",
    ):
        assert (
            get_default_backend(other, allow_multiple_backends_per_process=False)
            == default_backend
        )



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from pytest import CaptureFixture

import fedrq.cli


def test_cli_make_cache(patch_config_dirs, capsys: CaptureFixture):
    fedrq.cli.main(["make-cache"])
    out, err = capsys.readouterr()
    assert out == "Loaded 1 repo\n"
    assert not err



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Provides access to the default package backend interface.
It is recommended to use the backend directly instead of this module.
"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # Default to dnf backend for type checking
    from fedrq.backends.dnf import backend
    from fedrq.backends.dnf.backend import (
        BACKEND,
        BaseMaker,
        NEVRAForms,
        Repoquery,
        get_changelogs,
        get_releasever,
    )
else:
    from fedrq.backends import get_default_backend
    from fedrq.backends.base import (
        BackendMod,
        BaseMakerBase,
        NEVRAFormsCompat,
        PackageCompat,
        PackageQueryCompat,
        RepoqueryBase,
        _get_changelogs,
    )

    backend: BackendMod = get_default_backend()
    BaseMaker: type[BaseMakerBase] = backend.BaseMaker
    Package: type[PackageCompat] = backend.Package
    NEVRAForms: type[NEVRAFormsCompat] = backend.NEVRAForms
    PackageQuery: type[PackageQueryCompat] = backend.PackageQuery
    RepoError: type[BaseException] = backend.RepoError
    Repoquery: type[RepoqueryBase] = backend.Repoquery
    get_releasever: Callable[[], str] = backend.get_releasever
    get_changelogs: _get_changelogs = backend.get_changelogs
    BACKEND: str = backend.BACKEND

__all__ = (
    "BACKEND",
    "BaseMaker",
    "NEVRAForms",
    "Repoquery",
    "backend",
    "get_releasever",
    "get_changelogs",
)

warnings.warn(
    "The 'fedrq.repoquery' module is deprecated."
    " Import from the backend module directly"
    " or use `fedrq.backends.get_default_backend()`.",
    category=DeprecationWarning,
    stacklevel=2,
)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections.abc import Sequence

try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True

from fedrq.cli.base import CheckConfig, Command
from fedrq.cli.commands.cache import MakeCacheCommand
from fedrq.cli.commands.changelogs import ChangelogCommand
from fedrq.cli.commands.download import DownloadCommand, DownloadSpecCommand
from fedrq.cli.commands.formatters import FormattersCommand
from fedrq.cli.commands.pkgs import Pkgs
from fedrq.cli.commands.repolist import Repolist
from fedrq.cli.commands.subpkgs import Subpkgs
from fedrq.cli.commands.whatrequires import (
    WhatCommand,
    Whatenhances,
    Whatobsoletes,
    Whatrecommends,
    Whatrequires,
    WhatrequiresSrc,
    Whatsuggests,
    Whatsupplements,
)

__all__ = (
    "Command",
    "Pkgs",
    "Repolist",
    "Subpkgs",
    "ChangelogCommand",
    "DownloadCommand",
    "DownloadSpecCommand",
    "FormattersCommand",
    "MakeCacheCommand",
    "WhatCommand",
    "Whatenhances",
    "Whatobsoletes",
    "Whatrecommends",
    "Whatrequires",
    "WhatrequiresSrc",
    "Whatsuggests",
    "Whatsupplements",
)


def version() -> str:
    from fedrq import __version__

    return __version__


def main(argv: Sequence | None = None, **kwargs) -> None:
    parser = argparse.ArgumentParser(
        description="fedrq is a tool for querying the Fedora and EPEL repositories.",
        **kwargs,
    )
    parser.add_argument("--version", action="version", version=version())
    subparsers = parser.add_subparsers(
        title="Subcommands", dest="action", required=True
    )
    for name, cls in COMMANDS.items():
        cls.make_parser(subparsers.add_parser, name=name, add_help=True)
    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    return COMMANDS[args.action](args).run()


COMMANDS: dict[str, type[Command]] = {
    "make-cache": MakeCacheCommand,
    "check-config": CheckConfig,
    "changelog": ChangelogCommand,
    "download": DownloadCommand,
    "download-spec": DownloadSpecCommand,
    "formatters": FormattersCommand,
    "pkgs": Pkgs,
    "subpkgs": Subpkgs,
    "repolist": Repolist,
    "whatenhances": Whatenhances,
    "whatobsoletes": Whatobsoletes,
    "whatrecommends": Whatrecommends,
    "whatrequires": Whatrequires,
    "wr": Whatrequires,
    "whatrequires-src": WhatrequiresSrc,
    "wrsrc": WhatrequiresSrc,
    "whatsuggests": Whatsuggests,
    "whatsupplements": Whatsupplements,
}



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import fedrq.config


def test_source_repos_formatter():
    config = fedrq.config.get_config()
    release = config.get_release("f37", "@repo:fedora")
    bm = config.backend_mod.BaseMaker()
    bm.load_release_repos(release)
    assert bm.repolist(True) == ["fedora"]
    release.release_config.repogs.get_repo("@source-repos").load(bm, config, release)
    assert sorted(bm.repolist(True)) == sorted(["fedora", "fedora-source"])



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
fedrq is a tool to query the Fedora and EPEL repositories
"""

from __future__ import annotations

import logging
import os
import warnings

import pydantic

__version__ = "1.1.0"


def _filter_pydantic_v2_warnings() -> None:
    typ: type[DeprecationWarning] | None
    if typ := getattr(pydantic, "PydanticDeprecatedSince20", None):
        warnings.simplefilter(action="ignore", category=typ)


if "_FEDRQ_SHOW_PYDANTIC_WARNINGS" not in os.environ:
    _filter_pydantic_v2_warnings()


fmt = "{levelname}:{name}:{lineno}: {message}"
logging.basicConfig(format=fmt, style="{")
logger = logging.getLogger("fedrq")

__all__ = ("__version__", "logger")



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Test libdnf5-specific backend code
"""

from __future__ import annotations

import pytest

from fedrq.backends.base import BackendMod


@pytest.fixture(autouse=True)
def skip_mod(default_backend: BackendMod):
    if default_backend.BACKEND != "libdnf5":
        pytest.skip("This test checks libdnf5 functionality")


def test_libdnf5_bm_load_filelists():
    import fedrq.backends.libdnf5.backend as b

    bm = b.BaseMaker()
    default_types = sorted(bm.conf.optional_metadata_types)
    assert "filelists" not in default_types
    bm.load_filelists(False)
    assert sorted(bm.conf.optional_metadata_types) == default_types
    bm.load_filelists(True)
    new = sorted((*default_types, "filelists"))
    assert sorted(bm.conf.optional_metadata_types) == new



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import json
from collections.abc import Callable, Collection
from functools import cache
from pathlib import Path

import pytest

from fedrq import config as rqconfig
from fedrq.backends.base import PackageQueryCompat, RepoqueryBase
from fedrq.cli import formatters


@cache
def get_rq():
    return rqconfig.get_config(load_filelists="always").get_rq("tester", "base")


def formatter(
    query,
    formatter_name="plain",
    *args,
    attr=False,
    sort=True,
    repoquery: RepoqueryBase | None = None,
    **kwargs,
):
    result = [
        str(i)
        for i in formatters.DefaultFormatters.get_formatter(
            formatter_name, repoquery=repoquery
        ).format(query, *args, **kwargs)
    ]
    if sort:
        result.sort()
    if attr:
        result2 = [
            str(i)
            for i in formatters.DefaultFormatters.get_formatter(
                f"attr:{formatter_name}"
            ).format(query, *args, **kwargs)
        ]
        if sort:
            result2.sort()
        assert result == result2
    return result


# @pytest.mark.parametrize("special_repos", ("repo1",), indirect=["special_repos"])
def test_plain_formatter(patch_config_dirs, target_cpu):
    repo_test_rq = get_rq()
    expected = sorted(
        (
            "packagea-1-1.fc36.noarch",
            "packagea-1-1.fc36.src",
            "packagea-sub-1-1.fc36.noarch",
            "packageb-1-1.fc36.src",
            f"packageb-1-1.fc36.{target_cpu}",
            "packageb-11111:2-1.fc36.src",
            f"packageb-11111:2-1.fc36.{target_cpu}",
            "packageb-sub-1-1.fc36.noarch",
            "packageb-sub-11111:2-1.fc36.noarch",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query) == expected
    assert formatter(query, "plain") == expected


def test_plainwithrepo_formatter(patch_config_dirs, repo_test_rq, target_cpu):
    repo_test_rq = get_rq()
    expected = sorted(
        (
            "packagea-1-1.fc36.noarch testrepo1",
            "packagea-1-1.fc36.src testrepo1",
            "packagea-sub-1-1.fc36.noarch testrepo1",
            "packageb-1-1.fc36.src testrepo1",
            f"packageb-1-1.fc36.{target_cpu} testrepo1",
            "packageb-11111:2-1.fc36.src testrepo1",
            f"packageb-11111:2-1.fc36.{target_cpu} testrepo1",
            "packageb-sub-1-1.fc36.noarch testrepo1",
            "packageb-sub-11111:2-1.fc36.noarch testrepo1",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query, "plainwithrepo") == expected
    assert formatter(query, "nevrr") == expected


def test_name_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    expected = sorted(
        (
            "packagea",
            "packagea",
            "packagea-sub",
            "packageb",
            "packageb",
            "packageb",
            "packageb",
            "packageb-sub",
            "packageb-sub",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query, "name") == expected
    assert formatter(query, "attr:name") == expected


def test_evr_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name__glob="packageb*")
    result = sorted(
        (
            "11111:2-1.fc36",
            "11111:2-1.fc36",
            "11111:2-1.fc36",
            "1-1.fc36",
            "1-1.fc36",
            "1-1.fc36",
        )
    )
    assert formatter(query, "evr") == result
    assert formatter(query, "attr:evr") == result


def test_nv_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name__glob="packagea*")
    expected = sorted(("packagea-1", "packagea-1", "packagea-sub-1"))
    assert formatter(query, "nv") == expected


def test_source_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query()
    assert formatter(query, "source") == ["packagea", "packageb"]


@pytest.mark.parametrize(
    "latest,expected",
    (
        (None, ["1", "1", "2", "2"]),
        (1, ["2", "2"]),
    ),
)
def test_version_formatter(patch_config_dirs, latest, expected):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name="packageb", latest=latest)
    assert formatter(query, "version") == expected
    assert formatter(query, "attr:version") == expected


def test_epoch_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name="packageb-sub")
    assert formatter(query, "epoch") == ["0", "11111"]
    assert formatter(query, "attr:epoch") == ["0", "11111"]


def test_requires_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name=("packagea-sub", "packageb-sub"))
    assert len(query) == 3
    expected = [
        "/usr/share/packageb-sub",
        "package(b)",
        "packagea = 1-1.fc36",
        "vpackage(b) = 1-1.fc36",
    ]
    assert formatter(query, "requires", attr=True) == expected


def test_repo_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query()
    result = formatter(query, "reponame", attr=True)
    assert len(query) == len(result)
    assert {"testrepo1"} == set(result)


def test_repo_license_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name__glob="packagea*")
    result = formatter(query, "license", attr=True)
    assert result == ["Unlicense"] * 3


def test_debug_name_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name="packageb")
    result = formatter(query, "debug_name", attr=True)
    assert result == ["packageb-debuginfo"] * len(query)


def test_repo_files_formatter(patch_config_dirs):
    repo_test_rq = get_rq()
    query = repo_test_rq.query(name=["packagea", "packageb"], arch="notsrc", latest=1)
    result = formatter(query, "files", attr=True)
    assert result == ["/usr/share/packagea", "/usr/share/packageb"]


@pytest.mark.parametrize(
    "args, expected_call",
    [
        pytest.param([], lambda x: x),
        pytest.param([None], lambda x: x),
        pytest.param([{"file"}], lambda x: x),
        pytest.param([{"ftp"}], lambda _: None),
    ],
)
def test_remote_location(
    patch_config_dirs,
    data_path: Path,
    args: list[Collection[str] | None],
    expected_call: Callable[[str], str | None],
):
    path = (
        data_path
        / "repos"
        / "repo1"
        / "repo"
        / "SRPMS"
        / "specs"
        / "packagea-1-1.fc36.src.rpm"
    )
    expected = expected_call(f"file://{path}")

    repo_test_rq = get_rq()
    package = repo_test_rq.get_package("packagea", "src")
    result = package.remote_location(*args)
    assert result == expected


def test_formatter_remote_location(patch_config_dirs, data_path: Path):
    path = (
        data_path
        / "repos"
        / "repo1"
        / "repo"
        / "SRPMS"
        / "specs"
        / "packagea-1-1.fc36.src.rpm"
    )
    expected = f"file://{path}"

    repo_test_rq = get_rq()
    package = repo_test_rq.get_package("packagea", "src")
    result = formatter([package], "remote_location")
    assert result == [expected]


@pytest.mark.parametrize("attr", formatters._ATTRS)
def test_formatter_sanity(patch_config_dirs, attr):
    """
    Sanity test to ensure that supported formatters work at all
    """
    repo_test_rq = get_rq()
    query = repo_test_rq.query(
        name=["packagea", "packagea-sub", "packageb", "packageb-sub"], latest=1
    )
    result = formatter(query, attr)
    if attr not in (
        "provides",
        "requires",
        "obsoletes",
        "conflicts",
        "recommends",
        "suggests",
        "enhances",
        "supplements",
    ):
        assert len(result) == len(query)


def test_json_formatter(patch_config_dirs):
    expected = [
        {
            "name": "packagea",
            "evr": "1-1.fc36",
            "arch": "noarch",
            "requires": ["vpackage(b)"],
            "conflicts": [],
            "provides": ["package(a)", "packagea = 1-1.fc36", "vpackage(a) = 1-1.fc36"],
            "source_name": "packagea",
        },
        {
            "name": "packagea",
            "evr": "1-1.fc36",
            "arch": "src",
            "requires": ["vpackage(b) > 0"],
            "conflicts": [],
            "provides": ["packagea = 1-1.fc36", "packagea-sub = 1-1.fc36"],
            "source_name": None,
        },
        {
            "name": "packagea-sub",
            "evr": "1-1.fc36",
            "arch": "noarch",
            "requires": ["/usr/share/packageb-sub", "packagea = 1-1.fc36"],
            "conflicts": [],
            "provides": [
                "subpackage(a)",
                "packagea-sub = 1-1.fc36",
                "vsubpackage(a) = 1-1.fc36",
            ],
            "source_name": "packagea",
        },
    ]
    repo_test_rq = get_rq()
    query = repo_test_rq.resolve_pkg_specs(["packagea*"], latest=1)
    output = formatter(
        query, "json:name,evr,arch,requires,conflicts,provides,source_name"
    )
    assert len(output) == 1
    assert json.loads(output[0]) == expected


def test_multiline_formatter(patch_config_dirs, target_cpu: str):
    repo_test_rq = get_rq()
    query = repo_test_rq.resolve_pkg_specs(
        ["packagea-1", "packageb"], with_src=False, latest=1
    )
    output = formatter(query, "multiline:na,description", sort=False)
    assert output == [
        "packagea.noarch : packagea is a test package.",
        "packagea.noarch : This is another line of text.",
        "packagea.noarch : Another another.",
        "packagea.noarch : And another.",
        f"packageb.{target_cpu} : ...",
    ]


@cache
def formatter_test_query() -> PackageQueryCompat:
    repo_test_rq = get_rq()
    return repo_test_rq.resolve_pkg_specs(
        ["packagea-1", "packageb"], with_src=False, latest=1
    )


@pytest.mark.parametrize(
    "formatter_,expected_output",
    [
        pytest.param(
            "multiline:na,description",
            [
                "packagea.noarch : packagea is a test package.",
                "packagea.noarch : This is another line of text.",
                "packagea.noarch : Another another.",
                "packagea.noarch : And another.",
                "packageb.{target_cpu} : ...",
            ],
            id="multiline",
        ),
        pytest.param(
            "line:na,repoid",
            [
                "packagea.noarch : testrepo1",
                "packageb.{target_cpu} : testrepo1",
            ],
            id="line-simple",
        ),
        pytest.param(
            "line:na,repoid:",
            [
                "packagea.noarch : testrepo1",
                "packageb.{target_cpu} : testrepo1",
            ],
            id="line-trailing",
        ),
        pytest.param(
            "line:na,repoid: | ",
            [
                "packagea.noarch | testrepo1",
                "packageb.{target_cpu} | testrepo1",
            ],
            id="line-custom-separator",
        ),
        pytest.param(
            "line:na,repoid,source: | ",
            [
                "packagea.noarch | testrepo1 | packagea",
                "packageb.{target_cpu} | testrepo1 | packageb",
            ],
            id="line-special-formatter",
        ),
        pytest.param(
            "line:line:na,repoid: | ",
            [
                "packagea.noarch | testrepo1",
                "packageb.{target_cpu} | testrepo1",
            ],
            id="line-stacked",
        ),
        pytest.param(
            "description",
            [
                "packagea is a test package.\n"
                "This is another line of text.\nAnother another.\nAnd another.\n---\n",
                "...",
            ],
            id="description",
        ),
        pytest.param(
            [
                "requiresmatch:packageb",
                "rm:packageb",
                "rmsrc:packageb",
                "requiresmatch-src:packageb",
            ],
            ["vpackage(b)"],
            id="rm",
        ),
        pytest.param(
            ["narm:packageb", "na-requiresmatch:packageb,packagea,jfjfjfj"],
            ["packagea.noarch : vpackage(b)"],
            id="narm",
        ),
        pytest.param(
            ["source+requiresmatch:packageb", "source+rm:packageb"],
            ["packagea : vpackage(b)"],
            id="source+requiresmatch",
        ),
    ],
)
def test_formatter_p(
    patch_config_dirs,
    target_cpu: str,
    formatter_: str | Collection[str],
    expected_output: Collection[str],
) -> None:
    formatters = [formatter_] if isinstance(formatter_, str) else formatter_
    query = formatter_test_query()
    for fmt in formatters:
        output = formatter(query, fmt, sort=False, repoquery=get_rq())
        assert output == [
            item.format(target_cpu=target_cpu) for item in expected_output
        ], fmt



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later OR MIT

from __future__ import annotations

import os
from collections.abc import Sequence
from glob import iglob
from pathlib import Path
from shutil import copy2

import nox

IN_CI = "JOB_ID" in os.environ or "CI" in os.environ
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)
PINNED = os.environ.get("PINNED", "true").lower() in (
    "1",
    "true",
)

PROJECT = "fedrq"
SPECFILE = "fedrq.spec"
LINT_SESSIONS = ("formatters", "codeqa", "typing")
LINT_FILES = (f"src/{PROJECT}", "tests/", "noxfile.py")

nox.options.sessions = ("lint", "covtest")
nox.options.reuse_existing_virtualenvs = True


# Helpers


def install(
    session: nox.Session, *args, editable=False, constraint: str | None = None, **kwargs
):
    # nox --no-venv
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn(f"No venv. Skipping installation of {args}")
        return
    largs = []
    if constraint and PINNED:
        largs.extend(("-c", f"requirements/{constraint}.txt"))
    if editable and ALLOW_EDITABLE:
        largs.append("-e")
    session.install(*largs, *args, **kwargs)


def git(session: nox.Session, *args, **kwargs):
    return session.run("git", *args, **kwargs, external=True)


# General


@nox.session(venv_params=["--system-site-packages"])
def test(
    session: nox.Session,
    backend: str | None = None,
    posargs: Sequence[str] | None = None,
):
    if not backend:
        # Make sure pytest is updated, as using the version from system
        # site-packages causes problems with plugins.
        install(session, ".[test]", constraint="test", editable=True)
    tmp = Path(session.create_tmp())
    env = {"FEDRQ_BACKEND": backend} if backend else {}
    if any(i.startswith("--cov") for i in session.posargs):
        install(session, "coverage[toml]", "pytest-cov")
        env |= {"COVERAGE_FILE": str(tmp / ".coverage")}
    session.run(
        "pytest", *(posargs if posargs is not None else session.posargs), env=env
    )


@nox.session
def coverage(session: nox.Session):
    install(session, "coverage[toml]")
    session.run("coverage", "combine", "--keep", *iglob(".nox/*test/tmp/.coverage"))
    session.run("coverage", "html")
    session.run("coverage", "report", "--fail-under=90")


@nox.session
def covtest(session: nox.Session):
    session.run("rm", "-f", *iglob(".nox/*/tmp/.coverage"), external=True)
    for name in ("dnf_test", "libdnf5_test", "pydanticv1_test", "coverage"):
        session.notify(name, [*session.posargs, "--cov"])


@nox.session(venv_backend="none")
def lint(session: nox.Session):
    """
    Run formatters, codeql, typing, and reuse sessions
    """
    for notify in LINT_SESSIONS:
        session.notify(notify)


@nox.session()
def codeqa(session: nox.Session):
    install(session, ".[codeqa]", constraint="codeqa")
    session.run("ruff", "check", *session.posargs, *LINT_FILES)
    session.run("reuse", "lint")


@nox.session
def formatters(session: nox.Session):
    install(session, ".[formatters]", constraint="formatters")
    posargs = session.posargs
    if IN_CI:
        posargs.append("--check")
    session.run("black", *posargs, *LINT_FILES)
    session.run("isort", *posargs, *LINT_FILES)


@nox.session
def typing(session: nox.Session):
    install(session, ".[typing]", editable=True, constraint="typing")
    session.run("mypy", "--enable-incomplete-feature=Unpack", "src/fedrq/")


@nox.session(reuse_venv=False)
def bump(session: nox.Session):
    version = session.posargs[0]

    install(session, "releaserr", "flit", "fclogr")

    session.run("releaserr", "check-tag", version)
    session.run("releaserr", "ensure-clean")
    session.run("releaserr", "set-version", "-s", "file", version)

    install(session, ".")

    # Bump specfile
    # fmt: off
    session.run(
        "fclogr", "bump",
        "--new", version,
        "--comment", f"Release {version}.",
        SPECFILE,
    )
    # fmt: on

    # Bump changelog, commit, and tag
    git(session, "add", SPECFILE, f"src/{PROJECT}/__init__.py")
    session.run("releaserr", "clog", version, "--tag")
    session.run("releaserr", "build", "--sign", "--backend", "flit_core")


@nox.session(reuse_venv=False)
def publish(session: nox.Session):
    # Setup
    install(session, "releaserr", "twine")
    session.run("releaserr", "ensure-clean")

    # Upload to PyPI
    session.run("releaserr", "upload")

    # Push to git, publish artifacts to sourcehut, and release to copr
    if not session.interactive or input(
        "Push to Sourcehut and copr build (Y/n)"
    ).lower() in ("", "y"):
        git(session, "push", "--follow-tags")
        session.run("hut", "git", "artifact", "upload", *iglob("dist/*"), external=True)
        copr_release(session)

    # Post-release bump
    session.run("releaserr", "post-version", "-s", "file")
    git(session, "add", f"src/{PROJECT}/__init__.py")
    git(session, "commit", "-S", "-m", "Post release version bump")


@nox.session
def copr_release(session: nox.Session):
    install(session, "copr-cli", "requests-gssapi", "specfile")
    tmp = Path(session.create_tmp())
    dest = tmp / SPECFILE
    copy2(SPECFILE, dest)
    session.run("python", "contrib/fedoraify.py", str(dest))
    session.run("copr-cli", "build", "--nowait", f"gotmax23/{PROJECT}", str(dest))


@nox.session
def srpm(session: nox.Session, posargs=None):
    install(session, "-r", "requirements/srpm.in", constraint="srpm")
    posargs = posargs or session.posargs
    session.run("python3", "-m", "fclogr", "--debug", "dev-srpm", *posargs)


@nox.session
def mockbuild(session: nox.Session):
    tmp = Path(session.create_tmp())
    srpm(session, ("-o", tmp, "--keep"))
    spec_path = tmp / "fedrq.spec"
    margs = [
        "mock",
        "--spec",
        str(spec_path),
        "--source",
        str(tmp),
        *session.posargs,
    ]
    if not session.interactive:
        margs.append("--verbose")
    session.run(*margs, external=True)


# fedrq specific


@nox.session
def mkdocs(session: nox.Session):
    install(session, "-e", ".[doc]", constraint="doc")
    session.run("mkdocs", *(session.posargs or ["build"]))


@nox.session(venv_backend="none")
def testa(session: nox.Session):
    session.notify("dnf_test", ["--cov"])
    session.notify("libdnf5_test", ["--cov"])
    session.notify("pydanticv1_test")
    session.notify("coverage")


@nox.session(venv_params=["--system-site-packages"])
def dnf_test(session: nox.Session):
    install(session, ".[test]", constraint="test", editable=True)
    test(session, "dnf")


@nox.session
def libdnf5_test(session: nox.Session):
    install(session, ".[test]", "libdnf5-shim", constraint="test", editable=True)
    test(session, "libdnf5")


@nox.session(venv_params=["--system-site-packages"])
def pydanticv1_test(session: nox.Session):
    install(session, ".[test]", constraint="pydanticv1_test")
    test(session, "dnf", ["tests/unit"])


@nox.session(
    name="pip-compile", python=["3.9"], venv_params=["--download"], reuse_venv=False
)
def pip_compile(session: nox.Session):
    session.install("pip-tools")
    Path("requirements").mkdir(exist_ok=True)

    # Use --upgrade by default unless a user passes -P.
    args = list(session.posargs)
    if not any(
        arg.startswith(("-P", "--upgrade-package", "--no-upgrade")) for arg in args
    ):
        args.append("--upgrade")

    # fmt: off
    session.run(
        "pip-compile",
        "-o", "requirements/requirements.txt",
        *args,
    )

    extras = (
        "codeqa",
        "doc",
        "formatters",
        "typing",
        "test",
    )
    for extra in extras:
        session.run(
            "pip-compile",
            "-o", f"requirements/{extra}.txt",
            f"--extra={extra}",
            *args,
        )

    extras_a = [f"--extra={extra}" for extra in extras]
    session.run("pip-compile", "-o", "requirements/all.txt", *extras_a, *args)

    session.run(
        "pip-compile", "-o", "requirements/srpm.txt", *args, "requirements/srpm.in"
    )

    session.run(
        "pip-compile",
        "-o", "requirements/pydanticv1_test.txt",
        "-c", "requirements/pydanticv1.in",
        "--extra=test",
        *args,
    )
    # fmt: on



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections.abc import Callable

from fedrq._utils import mklog
from fedrq.cli.base import Command
from fedrq.cli.formatters import DefaultFormatters


class FormattersCommand(Command):
    """
    List available formatters
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args

        self.v_logging()
        flog = mklog(__name__, self.__class__.__name__)
        flog.debug("args=%s", args)

    @classmethod
    def make_parser(
        cls,
        parser_func: Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        parser = super().make_parser(
            parser_func, add_help=add_help, parents=[], **kwargs
        )
        parser.add_argument(
            "--only-formatters", "--formatters", action="store_true", dest="formatters"
        )
        parser.add_argument(
            "--only-attrs", "--attrs", action="store_true", dest="attrs"
        )
        parser.add_argument(
            "--only-special-formatters",
            "--special-formatters",
            action="store_true",
            dest="special_formatters",
        )
        return parser

    def run(self) -> None:
        formatter_args_names = {"formatters", "attrs", "special_formatters"}
        formatter_args = {
            key: value
            for key, value in vars(self.args).items()
            if key in formatter_args_names
        }
        # If none of the --only-foo args are passed, set everything to True
        if set(formatter_args.values()) == {False}:
            formatter_args = dict.fromkeys(formatter_args.keys(), True)
        for formatter in sorted(DefaultFormatters.formatters_it(**formatter_args)):
            print(formatter)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import abc
import dataclasses
import importlib.resources
import logging
from collections.abc import Callable, Collection, Iterable, Iterator
from datetime import date
from typing import TYPE_CHECKING, Any, Generic, Optional, Protocol, TypeVar
from warnings import warn

if TYPE_CHECKING:
    from _typeshed import StrPath

    from fedrq.config import Release

_QueryT = TypeVar("_QueryT", bound="PackageQueryCompat")
_PackageT = TypeVar("_PackageT", bound="PackageCompat")
LOG = logging.getLogger("fedrq.backends")


class PackageCompat(metaclass=abc.ABCMeta):  # pragma: no cover
    """
    Common interface provided by dnf.package.Package and other backends
    """

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @property
    @abc.abstractmethod
    def arch(self) -> str: ...

    @property
    @abc.abstractmethod
    def a(self) -> str: ...

    @property
    @abc.abstractmethod
    def epoch(self) -> int: ...

    @property
    @abc.abstractmethod
    def e(self) -> int: ...

    @property
    @abc.abstractmethod
    def version(self) -> str: ...

    @property
    @abc.abstractmethod
    def v(self) -> str: ...

    @property
    @abc.abstractmethod
    def release(self) -> str: ...

    @property
    @abc.abstractmethod
    def r(self) -> str: ...

    @property
    @abc.abstractmethod
    def from_repo(self) -> str: ...

    @property
    @abc.abstractmethod
    def evr(self) -> str: ...

    @property
    @abc.abstractmethod
    def debug_name(self) -> str: ...

    @property
    @abc.abstractmethod
    def source_name(self) -> Optional[str]: ...

    @property
    @abc.abstractmethod
    def source_debug_name(self) -> str: ...

    @property
    @abc.abstractmethod
    def installtime(self) -> int: ...

    @property
    @abc.abstractmethod
    def buildtime(self) -> int: ...

    @property
    @abc.abstractmethod
    def size(self) -> int: ...

    @property
    @abc.abstractmethod
    def downloadsize(self) -> int: ...

    @property
    @abc.abstractmethod
    def installsize(self) -> int: ...

    @property
    @abc.abstractmethod
    def provides(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def requires(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def recommends(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def suggests(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def supplements(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def enhances(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def obsoletes(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def conflicts(self) -> Iterable: ...

    @property
    @abc.abstractmethod
    def sourcerpm(self) -> Optional[str]: ...

    @property
    @abc.abstractmethod
    def description(self) -> str: ...

    @property
    @abc.abstractmethod
    def summary(self) -> str: ...

    @property
    @abc.abstractmethod
    def license(self) -> str: ...

    @property
    @abc.abstractmethod
    def url(self) -> str: ...

    @property
    @abc.abstractmethod
    def reason(self) -> Optional[str]: ...

    @property
    @abc.abstractmethod
    def files(self) -> Iterable[str]: ...

    @property
    @abc.abstractmethod
    def reponame(self) -> str: ...

    @property
    @abc.abstractmethod
    def repoid(self) -> str: ...

    @property
    @abc.abstractmethod
    def vendor(self) -> str: ...

    @property
    @abc.abstractmethod
    def packager(self) -> str: ...

    @property
    @abc.abstractmethod
    def location(self) -> str: ...

    @property
    @abc.abstractmethod
    def repo(self) -> Any:
        """
        Return the package's Repo object.
        The exact object depends on which backend is used.
        """

    @abc.abstractmethod
    def remote_location(
        self, schemes: Collection[str] | None = ("http", "ftp", "file", "https")
    ) -> str | None: ...

    @abc.abstractmethod
    def __hash__(self) -> int: ...

    @abc.abstractmethod
    def __lt__(self, other) -> bool: ...

    @abc.abstractmethod
    def __le__(self, other) -> bool: ...

    @abc.abstractmethod
    def __gt__(self, other) -> bool: ...

    @abc.abstractmethod
    def __ge__(self, other) -> bool: ...


class PackageQueryCompat(Generic[_PackageT], metaclass=abc.ABCMeta):  # pragma: no cover
    """
    Common PackageQuery interface provided by hawkey.Query and other backends.
    """

    @abc.abstractmethod
    def filter(self: _QueryT, **kwargs) -> _QueryT:
        """
        Filter the PackageQuery.
        Depending on the backend, this either modifies 'self' in place and
        return 'self' or return a new PackageQuery object.
        See https://dnf.readthedocs.io/en/latest/api_queries.html#dnf.query.Query.filter
        for the allowed kwargs.
        """
        ...

    @abc.abstractmethod
    def filterm(self: _QueryT, **kwargs) -> _QueryT:
        """
        Filter the PackageQuery in place and return 'self'.
        See https://dnf.readthedocs.io/en/latest/api_queries.html#dnf.query.Query.filter
        for the allowed kwargs.
        """
        ...

    @abc.abstractmethod
    def union(self: _QueryT, other: _QueryT) -> _QueryT:
        """
        Combine two PackageQuery objects.
        Depending on the backend, this either modifies 'self' in place and
        returns 'self' or returns a new PackageQuery object.
        """
        ...

    @abc.abstractmethod
    def __len__(self) -> int: ...

    @abc.abstractmethod
    def __iter__(self) -> Iterator[_PackageT]: ...


class BaseMakerBase(metaclass=abc.ABCMeta):
    """
    Create a Base object, set configuration, and load repos
    """

    base: Any

    def __init__(self, base=None) -> None:
        self.base = base

    @property
    @abc.abstractmethod
    def conf(self) -> Any:
        """
        Return the backend's Config object
        """
        ...

    @abc.abstractmethod
    def fill_sack(
        self,
        *,
        from_cache: bool = False,
        load_system_repo: bool = False,
    ) -> Any:
        """
        Fill the sack and returns the Base object.
        The repository configuration shouldn't be manipulated after this.
        'from_cache' isn't currently supported by the libdnf5 backend.
        """
        ...

    @abc.abstractmethod
    def read_system_repos(self, disable: bool = True) -> None:
        """
        Load system repositories into the base object.
        By default, they are all disabled even if 'enabled=1' is in the
        repository configuration.
        """

    @abc.abstractmethod
    def enable_repos(self, repos: Collection[str]) -> None:
        """
        Enable a list of repositories by their repoid.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """

    @abc.abstractmethod
    def enable_repo(self, repo: str) -> None:
        """
        Enable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """

    @abc.abstractmethod
    def disable_repo(self, repo: str, ignore_missing: bool = True) -> None:
        """
        Disable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration
        when ignore_missing is False.
        """

    @abc.abstractmethod
    def read_repofile(self, file: StrPath) -> None:
        """
        Load repositories from a repo file
        """

    @abc.abstractmethod
    def set(self, key: str, value: Any) -> None:
        """
        Set configuration options. Must be called before reading repos.
        """
        ...

    @abc.abstractmethod
    def set_var(self, key: str, value: Any) -> None:
        """
        Set substitutions (e.g. arch, basearch, releasever).
        Needs to be called before reading repos.
        """
        ...

    # Private for now
    @abc.abstractmethod
    def _read_repofile_new(self, file: StrPath, ensure_enabled: bool = False) -> None:
        """
        Load repositories from a repo file if they're not already in the
        configuration.
        """

    def sets(self, conf: dict[str, Any], substitutions: dict[str, Any]) -> None:
        """
        Set options on the base object

        Args:
            conf:
                A dict of configuration options. Call self.set() for each k-v
                pair.
            substitutions:
                A dict of substitutions/vars options. Call self.set_var() for
                each k-v pair.
        """
        for opt in conf.items():
            self.set(*opt)
        for opt in substitutions.items():
            self.set_var(*opt)

    def load_filelists(self, enable: bool = True) -> None:  # noqa: ARG002
        # Can be overriden by subclasses. Purposely isn't an @abstractmethod.
        """
        Load the filelists if they're not already enabled by default

        Args:
            enable:
                Whether to enable or disable filelists
        """
        return None

    @abc.abstractmethod
    def load_changelogs(self, enable: bool = True) -> None:
        """
        Load changelog metadata

        Args:
            enable:
                Whether to enable or disable filelists
        """

    def load_release_repos(self, release: Release, set_releasever: bool = True) -> None:
        """
        Load the repositories from a fedrq.config.Release object

        Args:
            release:
                [`Release`][fedrq.config.Release] object
            set_releasever:
                Whether to set the `$releasever` based on the release or just
                leave it alone
        """
        if set_releasever:
            self.set_var("releasever", release.version)
        if release.release_config.system_repos:
            self.read_system_repos(
                disable=not release.release_config.append_system_repos
            )
        for path in release.release_config.full_def_paths:
            with importlib.resources.as_file(path) as fp:
                LOG.debug("Reading %s", fp)
                self._read_repofile_new(fp)
        release.repog.load(self, release.config, release)

    @abc.abstractmethod
    def create_repo(self, repoid: str, **kwargs: Any) -> None:
        """
        Add a Repo object to the repo sack and configure it.

        Args:
            repoid:
                Repository ID
            kwargs:
                key-values options that should be set on the Repo object values
                (like $basearch) will be substituted automatically.
        """
        ...

    @property
    @abc.abstractmethod
    def backend(self) -> BackendMod: ...

    @abc.abstractmethod
    def repolist(self, enabled: bool | None = None) -> list[str]: ...

    @abc.abstractmethod
    def enable_source_repos(self) -> None:
        """
        Enable the corresponding -source repos of the currently enabled
        repositories
        """
        ...


class NEVRAFormsCompat(Protocol):
    NEVRA: int
    NEVR: int
    NEV: int
    NA: int
    NAME: int


class RepoqueryBase(Generic[_QueryT], metaclass=abc.ABCMeta):
    """
    Helpers to query a repository.
    Provides a unified repoquery interface for different backends.
    """

    def __init__(self, base) -> None:
        self.base = base

    @property
    @abc.abstractmethod
    def base_arches(self) -> set[str]:
        """
        Return a set of the system's arch and basearch.
        """
        ...

    def _get_resolve_options(
        self,
        resolve: bool,
        with_filenames: bool | None,
        with_provides: bool | None,
        resolve_provides: bool | None,
    ) -> dict[str, Any]:
        opts: dict[str, bool | None] = {
            "with_filenames": with_filenames,
            "with_provides": with_provides,
            "resolve_provides": resolve_provides,
        }
        return {key: resolve if opt is None else opt for key, opt in opts.items()}

    @abc.abstractmethod
    def resolve_pkg_specs(
        self,
        specs: Collection[str],
        resolve: bool = False,
        latest: int | None = None,
        with_src: bool = True,
        *,
        with_filenames: bool | None = None,
        with_provides: bool | None = None,
        resolve_provides: bool | None = None,
    ) -> _QueryT:
        """
        Resolve pkg specs.
        See
        https://dnf.readthedocs.io/en/latest/command_ref.html?highlight=spec#specifying-packages
        or
        https://dnf5.readthedocs.io/en/latest/misc/specs.7.html
        for valid forms.

        Args:
            specs:
                Package specs to resolve.
            resolve:
                Whether to resolve file paths or virtual Provides in addition
                to package specs
            latest:
                Limit packages with the same name and arch.
            with_src:
                Whether to consider `.src` packages when resolving `specs`
        """
        ...

    def arch_filterm(
        self: RepoqueryBase[_QueryT],
        query: _QueryT,
        arch: str | Iterable[str] | None = None,
    ) -> _QueryT:
        """
        Filter a query's architectures in place and return it.
        It includes a little more functionality than query.filterm(arch=...).

        - When arch is None, the query is left untouched.
        - If arch equals 'notsrc', all src and multilib packages are
          excluded.
        - If arch equals 'arched', all noarch, multilib, and source
          packages are excluded.
        - Otherwise, arch is passed to query.filterm(arch=...) and no other
          validation is preformed.
        """
        if not arch:
            return query
        if arch == "notsrc":
            return query.filterm(arch=(*self.base_arches, "noarch"))  # type: ignore
        elif arch == "arched":
            return query.filterm(arch=self.base.conf.basearch)
        else:
            return query.filterm(arch=arch)

    def arch_filter(
        self: RepoqueryBase[_QueryT],
        query: _QueryT,
        arch: str | Iterable[str] | None = None,
    ) -> _QueryT:
        """
        Filter a query's architectures and return it.
        It includes a little more functionality than query.filter(arch=...).

        - When arch is None, the query is left untouched.
        - If arch equals 'notsrc', all src and multilib packages are
          excluded.
        - If arch equals 'arched', all noarch, multilib, and source
          packages are excluded.
        - Otherwise, arch is passed to query.filterm(arch=...) and no other
          validation is preformed.
        """
        if not arch:
            return query
        if arch == "notsrc":
            return query.filter(arch=(*self.base_arches, "noarch"))  # type: ignore
        if arch == "arched":
            return query.filter(arch=list(self.base_arches))
        return query.filter(arch=arch)

    @abc.abstractmethod
    def _query(self) -> _QueryT:
        """
        Return the PackageQuery object for this backend
        """
        return self.base.sack.query()

    def query(
        self,
        *,
        arch: str | Iterable[str] | None = None,
        **kwargs,
    ) -> _QueryT:
        """
        Return an inital PackageQuery that's filtered with **kwargs.
        Further filtering can be applied with the PackageQuery's filter and
        filterm methods.
        """
        if kwargs.get("latest") is None:
            kwargs.pop("latest", None)
        query = self._query()
        query.filterm(**kwargs)
        self.arch_filterm(query, arch)
        return query

    def get_package(
        self,
        name: str,
        arch: str | Iterable[str] | None = None,
    ) -> PackageCompat:
        """
        Return the latest Package that matches the 'name' and 'arch'.
        A ValueError is raised when no matches are found.
        """
        query = self.query(name=name, arch=arch, latest=1)
        if len(query) < 1:
            raise ValueError(f"Zero packages found for {name} on {arch}")
        return next(iter(query))

    def get_subpackages(self, packages: Iterable[PackageCompat], **kwargs) -> _QueryT:
        """
        Return a PackageQuery containing the binary RPMS/subpackages produced
        by {packages}.

        Args:
            packages:
                An interable of `PackageCompat` containing source packages
        """
        arch = kwargs.get("arch")
        if arch == "src":
            raise ValueError("{arch} cannot be 'src'")
        elif not arch:
            kwargs.setdefault("arch__neq", "src")
        if val := kwargs.pop("sourcerpm", None):
            warn(f"Removing invalid kwarg: 'sourcerpm={val}")

        for package in packages:
            if package.arch != "src":
                raise ValueError(f"{package} must be a source package.")

        sourcerpms = [
            f"{package.name}-{package.version}-{package.release}.src.rpm"
            for package in packages
        ]
        query = self.query(sourcerpm=sourcerpms, **kwargs)
        return query

    @property
    @abc.abstractmethod
    def backend(self) -> BackendMod: ...


@dataclasses.dataclass(frozen=True)
class ChangelogEntry:
    """
    Data class for changelog entry data.
    Do not instantiate directly!
    """

    text: str
    author: str
    date: date

    def __str__(self) -> str:
        date_str = format(self.date, "%a %b %d %Y")
        return f"* {date_str} {self.author}\n{self.text}"


class _get_changelogs(Protocol):
    def __call__(self, package: Any) -> Iterator[ChangelogEntry]:
        """
        Args:
            package:
                A backend's Package object
        """
        ...


class BackendMod(Protocol):
    """
    Protocol for a fedrq backend module.
    Each backend module (e.g.
    [`fedrq.backends.dnf.backend`][fedrq.backends.dnf.backend])
    implements this interface.
    """

    BACKEND: str
    BaseMaker: type[BaseMakerBase]
    Package: type[PackageCompat]
    NEVRAForms: type[NEVRAFormsCompat]
    PackageQuery: type[PackageQueryCompat]
    Repoquery: type[RepoqueryBase]
    RepoError: type[BaseException]
    get_releasever: Callable[[], str]
    get_changelogs: _get_changelogs



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path

from mkdocs_gen_files.editor import FilesEditor
from releaserr.scd import scd2md

HERE = Path(__file__).resolve().parent

editor = FilesEditor.current()


def main() -> None:
    files = list(HERE.glob("*.scd"))
    new_files: list[Path] = scd2md(
        files, Path(editor.directory), to="markdown_strict+pipe_tables+smart"
    )
    for file in new_files:
        editor._get_file(str(file.relative_to(editor.directory)), True)


main()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from pathlib import Path

import pytest

import fedrq.cli


@pytest.mark.no_rpm_mock
def test_download_spec(tmp_path: Path):
    fedrq.cli.main(
        [
            "download-spec",
            "ansible-core",
            "-b",
            "f38",
            "-r",
            "@release",
            "-y",
            "-o",
            str(tmp_path),
        ]
    )
    assert (tmp_path / "ansible-core.spec").exists()
